# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
from abc import ABC
from abc import abstractmethod


class Output(ABC):
    """_summary_"""

    def __init__(self, context):
        self.context = context
        self.checkpoint = context.checkpoint

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @abstractmethod
    def write_initial_state(self, state):
        pass

    @abstractmethod
    def write_state(self, state):
        pass

    def reduce(self, state):
        """Creates new state which is projection of original state on the last step in the multi-steps dimension."""
        reduced_state = state.copy()
        reduced_state["fields"] = {}
        for field, values in state["fields"].items():
            reduced_state["fields"][field] = values[-1, :]
        return reduced_state

    def close(self):
        pass



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import argparse
import logging
import os
from functools import cached_property

from ai_models.model import Model

from anemoi.inference.inputs.grib import GribInput
from anemoi.inference.outputs.grib import GribOutput
from anemoi.inference.runner import PRECISIONS as AUTOCAST
from anemoi.inference.runners.plugin import PluginRunner

LOG = logging.getLogger(__name__)


class FieldListInput(GribInput):
    """
    Handles earchkit-data fieldlists input fields.
    """

    def __init__(self, context, *, input_fields):
        super().__init__(context)
        self.input_fields = input_fields

    def create_input_state(self, *, date):
        return self._create_input_state(self.input_fields, variables=None, date=date)

    def load_forcings(self, *, variables, dates):
        return self._load_forcings(self.input_fields, variables=variables, dates=dates)

    def set_private_attributes(self, state, input_fields):
        input_fields = input_fields.order_by("valid_datetime")
        state["_grib_templates_for_output"] = {field.metadata("name"): field for field in input_fields}


class CallbackOutput(GribOutput):
    """Call ai-models write method"""

    def __init__(self, context, *, write, encoding=None):
        super().__init__(context, encoding=encoding, templates={"source": "templates"})
        self.write = write

    def write_message(self, message, *args, **kwargs):
        self.write(message, *args, **kwargs)


class AIModelPlugin(Model):

    expver = None

    def add_model_args(self, parser) -> None:
        """To be implemented in subclasses to add model-specific arguments to the parser.

        Parameters
        ----------
        parser : argparse.ArgumentParser
            An instance of the parser to add arguments to.
        """
        pass

    def parse_model_args(self, args):
        parser = argparse.ArgumentParser()

        parser.add_argument("--checkpoint", required=not hasattr(self, "download_files"))
        parser.add_argument(
            "--autocast",
            type=str,
            choices=sorted(AUTOCAST.keys()),
        )

        self.add_model_args(parser)

        args = parser.parse_args(args)
        args._checkpoint = args.checkpoint

        if args._checkpoint is None:
            args._checkpoint = os.path.join(self.assets, self.download_files[0])

        for k, v in vars(args).items():
            setattr(self, k, v)

        return parser

    @cached_property
    def runner(self):
        return PluginRunner(self._checkpoint, device=self.device)

    def run(self):
        if self.deterministic:
            self.torch_deterministic_mode()

        input = FieldListInput(self.runner, input_fields=self.all_fields)
        output = CallbackOutput(self.runner, write=self.write)

        input_state = input.create_input_state(date=self.start_datetime)

        output.write_initial_state(input_state)

        for state in self.runner.run(input_state=input_state, lead_time=self.lead_time):
            output.write_state(state)

        output.close()

    # Below are methods forwarded to the checkpoint

    @property
    def param_sfc(self):
        return self.runner.param_sfc

    @property
    def param_level_pl(self):
        return self.runner.param_level_pl

    @property
    def param_level_ml(self):
        return self.runner.param_level_ml

    @property
    def constant_fields(self):
        return self.runner.checkpoint.constants_from_input

    @property
    def grid(self):
        return self.runner.checkpoint.grid

    @property
    def area(self):
        return self.runner.checkpoint.area

    @property
    def lagged(self):
        return self.runner.lagged


model = AIModelPlugin



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import numpy as np

from ..output import Output
from . import create_output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("extract_lam")
class ExtractLamOutput(Output):
    """_summary_"""

    def __init__(self, context, *, output, points="cutout_mask"):
        super().__init__(context)
        if isinstance(points, str):
            mask = self.checkpoint.load_supporting_array(points)
            points = -np.sum(mask)  # This is the global, we want the lam

        self.points = points
        self.output = create_output(context, output)

    def __repr__(self):
        return f"ExtractLamOutput({self.points}, {self.output})"

    def write_initial_state(self, state):
        self.output.write_initial_state(self._apply_mask(state))

    def write_state(self, state):
        self.output.write_state(self._apply_mask(state))

    def _apply_mask(self, state):

        if self.points < 0:
            # This is the global, we want the lam
            self.points = state["latitudes"].size + self.points

        state = state.copy()
        state["fields"] = state["fields"].copy()
        state["latitudes"] = state["latitudes"][: self.points]
        state["longitudes"] = state["longitudes"][: self.points]

        for field in state["fields"]:
            data = state["fields"][field]
            if data.ndim == 1:
                data = data[: self.points]
            else:
                data = data[..., : self.points]
            state["fields"][field] = data

        return state

    def close(self):
        self.output.close()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json

from earthkit.data.utils.dates import to_datetime

from ..config import load_config
from ..inputs.mars import postproc
from ..runners.default import DefaultRunner
from . import Command


class RetrieveCmd(Command):
    """Used by prepml."""

    need_logging = False

    def add_arguments(self, command_parser):
        command_parser.description = self.__doc__
        command_parser.add_argument("config", type=str, help="Path to checkpoint")
        command_parser.add_argument("--defaults", action="append", help="Sources of default values.")
        command_parser.add_argument("--date", type=str, help="Date")
        command_parser.add_argument("--output", type=str, help="Output file")
        command_parser.add_argument("--staging-dates", type=str, help="Path to a file with staging dates")
        command_parser.add_argument("--extra", action="append", help="Additional request values. Can be repeated")
        command_parser.add_argument("--retrieve-fields-type", type=str, help="Type of fields to retrieve")
        command_parser.add_argument("overrides", nargs="*", help="Overrides.")

    def run(self, args):

        config = load_config(args.config, args.overrides, defaults=args.defaults)

        runner = DefaultRunner(config)

        # TODO: Move this to the runner

        variables = runner.checkpoint.variables_from_input(include_forcings=True)
        area = runner.checkpoint.area
        grid = runner.checkpoint.grid

        if args.retrieve_fields_type is not None:
            selected = set()

            for name, kinds in runner.checkpoint.variable_categories().items():
                if "computed" in kinds:
                    continue
                for kind in kinds:
                    if args.retrieve_fields_type.startswith(kind):  # PrepML adds an 's' to the type
                        selected.add(name)

            variables = sorted(selected)

        extra = postproc(grid, area)

        for r in args.extra or []:
            k, v = r.split("=")
            extra[k] = v

        if args.staging_dates:
            dates = []
            with open(args.staging_dates) as f:
                for line in f:
                    dates.append(to_datetime(line.strip()))
        else:
            date = to_datetime(args.date)
            dates = [date + h for h in runner.checkpoint.lagged]

        requests = []
        for r in runner.checkpoint.mars_requests(
            dates=dates,
            variables=variables,
            use_grib_paramid=config.use_grib_paramid,
        ):
            r = r.copy()
            r.update(extra)
            requests.append(r)

        with open(args.output, "w") as f:
            json.dump(requests, f, indent=4)


command = RetrieveCmd



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from anemoi.utils.registry import Registry

runner_registry = Registry(__name__)


def create_runner(context, config):
    return runner_registry.from_config(config, context)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


def test_inference():
    pass


if __name__ == "__main__":
    test_inference()
